package example.localplugin;

import gridsim.dcworms.DCWormsTags;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import schedframe.events.scheduling.SchedulingEvent;
import schedframe.events.scheduling.TaskFinishedEvent;
import schedframe.events.scheduling.TaskRequestedTimeExpiredEvent;
import schedframe.resources.ResourceStatus;
import schedframe.resources.computing.ComputingResource;
import schedframe.resources.computing.Processor;
import schedframe.resources.units.ProcessingElements;
import schedframe.resources.units.ResourceUnit;
import schedframe.resources.units.ResourceUnitName;
import schedframe.resources.units.StandardResourceUnitName;
import schedframe.scheduling.UsedResourcesList;
import schedframe.scheduling.manager.resources.ClusterResourceManager;
import schedframe.scheduling.manager.resources.ResourceManager;
import schedframe.scheduling.manager.tasks.JobRegistry;
import schedframe.scheduling.plan.SchedulingPlanInterface;
import schedframe.scheduling.plan.impl.SchedulingPlan;
import schedframe.scheduling.plugin.grid.ModuleList;
import schedframe.scheduling.queue.TaskQueue;
import schedframe.scheduling.queue.TaskQueueList;
import schedframe.scheduling.tasks.TaskInterface;
import schedframe.scheduling.tasks.WorkloadUnit;
import dcworms.schedframe.scheduling.Executable;

public class FCFSCPUFreqScalingClusterLocalPlugin extends BaseLocalSchedulingPlugin {

	List<Processor> allocatedCPUs;
	public FCFSCPUFreqScalingClusterLocalPlugin () {
		allocatedCPUs = new ArrayList<Processor>();
	}

	public SchedulingPlanInterface<?> schedule(SchedulingEvent event, TaskQueueList queues, JobRegistry jobRegistry,
			ResourceManager resManager, ModuleList modules) {

		ClusterResourceManager resourceManager = (ClusterResourceManager) resManager;
		SchedulingPlan plan = new SchedulingPlan();
		// our tasks are placed only in first queue (see
		// BaseLocalPlugin.placeJobsInQueues() method)
		TaskQueue q = queues.get(0);
		// chose the events types to serve.
		// Different actions for different events are possible.
		switch (event.getType()) {
		
		case START_TASK_EXECUTION:

			// check all tasks in queue
			for (int i = 0; i < q.size(); i++) {
				WorkloadUnit job = q.get(i);
				TaskInterface<?> task = (TaskInterface<?>) job;
				// if status of the tasks in READY
				if (task.getStatus() == DCWormsTags.READY) {

					Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(resourceManager, task);
					if (choosenResources  != null) {
						addToSchedulingPlan(plan, task, choosenResources);
						ProcessingElements pes = (ProcessingElements)choosenResources.get(StandardResourceUnitName.PE);
						List<Processor> processors =  new ArrayList<Processor>();
						for(ComputingResource res : pes){
							processors.add((Processor) res);
						}
						adjustFrequency(ResourceStatus.BUSY,processors);
					} 
				}
			}
			break;
			
		case TASK_FINISHED:
			TaskFinishedEvent finEvent = (TaskFinishedEvent) event;
			Executable exec = (Executable) jobRegistry.getExecutable(finEvent.getJobId(), finEvent.getTaskId());
			UsedResourcesList usedResourcesList = exec.getUsedResources();
			ProcessingElements pes = (ProcessingElements)usedResourcesList.getLast().getResourceUnits().get(StandardResourceUnitName.PE);
			List<Processor> processors =  new ArrayList<Processor>();
			for(ComputingResource res : pes){
				processors.add((Processor) res);
				allocatedCPUs.add((Processor) res);
			}
			adjustFrequency(ResourceStatus.FREE, processors);
			break;
			
		case TASK_REQUESTED_TIME_EXPIRED:
			TaskRequestedTimeExpiredEvent timExpEvent = (TaskRequestedTimeExpiredEvent) event;
			exec = (Executable) jobRegistry.getExecutable(timExpEvent.getJobId(), timExpEvent.getTaskId());
			usedResourcesList = exec.getUsedResources();
			pes = (ProcessingElements)usedResourcesList.getLast().getResourceUnits().get(StandardResourceUnitName.PE);
			processors =  new ArrayList<Processor>();
			for(ComputingResource res : pes){
				allocatedCPUs.remove((Processor) res);
			}
			// check all tasks in queue
			for (int i = 0; i < q.size(); i++) {
				WorkloadUnit job = q.get(i);
				TaskInterface<?> task = (TaskInterface<?>) job;
				// if status of the tasks in READY
				if (task.getStatus() == DCWormsTags.READY) {

					Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(resourceManager, task);
					if (choosenResources  != null) {
						addToSchedulingPlan(plan, task, choosenResources);
						pes = (ProcessingElements)choosenResources.get(StandardResourceUnitName.PE);
						processors =  new ArrayList<Processor>();
						for(ComputingResource res : pes){
							processors.add((Processor) res);
						}
						adjustFrequency(ResourceStatus.BUSY, processors);
					} 
				}
			}
			break;
		}
		return plan;
	}
	
	private Map<ResourceUnitName, ResourceUnit> chooseResourcesForExecution(
			ClusterResourceManager resourceManager, TaskInterface<?> task) {

		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>();

		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 1;
		}

		if (cpuRequest != 0) {
			List<ComputingResource> choosenResources = null;
			List<Processor> processors = resourceManager.getProcessors();
			processors.removeAll(allocatedCPUs);
			if (processors.size() < cpuRequest) {
				// log.warn("Task requires more cpus than is availiable in this resource.");
				return null;
			}

			choosenResources = new ArrayList<ComputingResource>();

			for (int i = 0; i < processors.size() && cpuRequest > 0; i++) {
				if (processors.get(i).getStatus() == ResourceStatus.FREE) {
					choosenResources.add(processors.get(i));
					cpuRequest--;
				}
			}
			if (cpuRequest > 0) {
				// log.info("Task " + task.getJobId() + "_" + task.getId() +
				// " requires more cpus than is availiable in this moment.");
				return null;
			}

			ProcessingElements result = new ProcessingElements();
			result.addAll(choosenResources);
			map.put(StandardResourceUnitName.PE, result);
		}
		return map;
	}

	private void adjustFrequency(ResourceStatus status, List<Processor> processors){
		switch(status){
		case BUSY:
			for(Processor cpu: processors){
				try {
					if(cpu.getPowerInterface().getSupportedPStates().containsKey("P0"))
						cpu.getPowerInterface().setPState("P0");
				} catch (NoSuchFieldException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			break;
		case FREE:
			for(Processor cpu: processors){
				try {
					if(cpu.getPowerInterface().getSupportedPStates().containsKey("P3"))
						cpu.getPowerInterface().setPState("P3");
				} catch (NoSuchFieldException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			break;
		}
	}
	
	public String getName() {
		return getClass().getName();
	}

}